@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,10 +23,12 @@
 # Most of the guts of this file have been moved into im.py, with only some
 # of the stuff thar uses images remaining.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *
 
-import renpy.display
-import renpy.text
+
+
+import renpy
 from renpy.display.render import render, Render
 
 import collections
@@ -96,7 +98,7 @@ def get_tag_method(tag, method):
     it returns None.
     """
 
-    ti =  images.get((tag,), None)
+    ti = images.get((tag,), None)
 
     if ti is None:
         return None
@@ -113,6 +115,9 @@ def check_image_attributes(tag, attributes):
     Otherwise, returns None.
     """
 
+    negative = tuple(i for i in attributes if i[:1] == "-")
+    attributes = [i for i in attributes if i[:1] != "-"]
+
     l = [ ]
 
     for attrs, d in image_attributes[tag].items():
@@ -125,29 +130,33 @@ def check_image_attributes(tag, attributes):
 
             chosen = ca(tag, remainder, None)
             if chosen is not None:
-                l.append(list(attrs) + list(chosen))
+                l.append(attrs + tuple(chosen))
 
         else:
 
             if not remainder:
                 l.append(attrs)
 
+    if negative:
+        negated = {i[1:] for i in negative}
+        l = [ i for i in l if not (negated & set(i)) ]
+
     # Check to see if there's an image that is exactly the one we want.
     for i in l:
         if len(i) == len(attributes):
-            return tuple(i)
+            return tuple(i + negative)
 
     if len(l) != 1:
         return None
 
-    return tuple(l[0])
+    return tuple(l[0] + negative)
 
 
 def get_ordered_image_attributes(tag, attributes=(), sort=None):
     """
     :doc: image_func
 
-    Returns a list of image tags, ordered in a way that makes sense to
+    Returns a list of image attributes, ordered in a way that makes sense to
     present to the user.
 
     `attributes`
@@ -156,9 +165,10 @@ def get_ordered_image_attributes(tag, attributes=(), sort=None):
         can be in a single image at the same time.)
 
     `sort`
-        If not None, the returned list of attributes is sorted. This is a function
-        that should be used as a tiebreaker.
-
+        If not None, the returned list of attributes is sorted. This is a
+        one-argument function that should be used as a tiebreaker - see
+        `this tutorial <https://docs.python.org/3/howto/sorting.html#key-functions>`_
+        for more information.
     """
 
     sequences = [ ]
@@ -229,7 +239,7 @@ def get_ordered_image_attributes(tag, attributes=(), sort=None):
 
     for attr in attrcount:
         if attr not in rv:
-            l.append((attrtotalpos[attr] / attrcount[attr], sort(attr), attr))
+            l.append((attrtotalpos[attr] // attrcount[attr], sort(attr), attr))
 
     l.sort()
     for i in l:
@@ -292,7 +302,7 @@ def get_registered_image(name):
     """
     :doc: image_func
 
-    If an image with the same name has been registered with renpy.register_image,
+    If an image with the same name has been :ref:`registered <defining-images>`,
     returns it. Otherwise, returns None.
     """
 
@@ -338,9 +348,10 @@ class ImageReference(renpy.display.core.Displayable):
         super(ImageReference, self).__init__(**properties)
 
         self.name = name
+        self.target = None # type: renpy.display.core.Displayable|None
 
-    def __unicode__(self):
-        return u"<ImageReference {!r}>".format(self.name)
+    def _repr_info(self):
+        return repr(self.name)
 
     def __hash__(self):
         return hash(self.name)
@@ -381,6 +392,8 @@ class ImageReference(renpy.display.core.Displayable):
             if renpy.config.debug:
                 raise Exception(msg)
 
+        target = None # typing
+
         args = [ ]
 
         while name:
@@ -405,15 +418,15 @@ class ImageReference(renpy.display.core.Displayable):
         try:
 
             a = self._args.copy(name=name, args=args)
-
             self.target = target._duplicate(a)
 
         except Exception as e:
 
-            if renpy.config.debug:
+            if renpy.config.raise_image_exceptions and (renpy.config.debug or renpy.config.developer):
                 raise
 
             error(str(e))
+            return False
 
         # Copy the old transform over.
         new_transform = self.target._target()
@@ -444,7 +457,6 @@ class ImageReference(renpy.display.core.Displayable):
                 rv.name = rv.name._duplicate(args)
 
         rv.find_target()
-        rv._duplicatable = rv.target._duplicatable
 
         return rv
 
@@ -453,7 +465,8 @@ class ImageReference(renpy.display.core.Displayable):
         if self.target is None:
             self.find_target()
 
-        self._duplicatable = self.target._duplicatable
+        self.target._unique()
+        self._duplicatable = False
 
     def _in_current_store(self):
 
@@ -471,7 +484,7 @@ class ImageReference(renpy.display.core.Displayable):
 
     def _handles_event(self, event):
         if self.target is None:
-            return False
+            self.find_target()
 
         return self.target._handles_event(event)
 
@@ -542,10 +555,10 @@ class DynamicImage(renpy.display.core.Displayable):
     nosave = [ 'raw_target' ]
 
     # The target that this image currently resolves to.
-    target = None
+    target = None # type: renpy.display.core.Displayable|None
 
     # The raw target that the image resolves to, before it has been parameterized.
-    raw_target = None
+    raw_target = None # type: renpy.display.core.Displayable|None
 
     # Have we been locked, so we never change?
     locked = False
@@ -553,30 +566,42 @@ class DynamicImage(renpy.display.core.Displayable):
     # The name used for hashing.
     hash_name = None
 
+    # A copy of the last scope given to this displayable, if the
+    # displayable uses a prefix.
+    scope = None
+
     def __init__(self, name, scope=None, **properties):
         super(DynamicImage, self).__init__(**properties)
 
         self.name = name
 
-        if isinstance(name, basestring) and ("[prefix_" in name):
-            self._duplicatable = True
+        self._uses_scope = False
+
+        if isinstance(name, basestring):
+            if ("[prefix_" in name):
+                self._duplicatable = True
+
+            if "[" in name.replace("[prefix_]", ""):
+                self._uses_scope = True
 
         if isinstance(name, list):
             for i in name:
                 if ("[prefix_" in i):
                     self._duplicatable = True
 
-        if scope is not None:
-            self.find_target(scope)
-            self._uses_scope = True
-        else:
+                if "[" in i.replace("[prefix_]", ""):
+                    self._uses_scope = True
+
+        if scope is None:
             self._uses_scope = False
+        else:
+            self.find_target(scope)
 
     def _scope(self, scope, update):
         return self.find_target(scope, update)
 
-    def __unicode__(self):
-        return u"DynamicImage {!r}".format(self.name)
+    def _repr_info(self):
+        return repr(self.name)
 
     def __hash__(self):
 
@@ -622,6 +647,13 @@ class DynamicImage(renpy.display.core.Displayable):
         if self.locked and (self.target is not None):
             return
 
+        if scope is not None:
+            if self._uses_scope and (self._duplicatable or self._args.prefix):
+                self.scope = dict(scope)
+
+        elif self._uses_scope:
+            scope = self.scope
+
         if self._args.prefix is None:
             if self._duplicatable:
                 prefix = self.style.prefix
@@ -634,7 +666,7 @@ class DynamicImage(renpy.display.core.Displayable):
             search = [ ]
             target = renpy.easy.dynamic_image(self.name, scope, prefix=prefix, search=search)
         except KeyError as ke:
-            raise Exception("In DynamicImage %r: Could not find substitution '%s'." % (self.name, unicode(ke.args[0])))
+            raise Exception("In DynamicImage %r: Could not find substitution '%s'." % (self.name, str(ke.args[0])))
         except Exception as e:
             raise Exception("In DynamicImage %r: %r" % (self.name, e))
 
@@ -656,12 +688,16 @@ class DynamicImage(renpy.display.core.Displayable):
         if not update:
             return True
 
-        self.raw_target = target
+        raw_target = target # type: renpy.display.core.Displayable
         old_target = self.target
 
-        if target._duplicatable:
-            target = target._duplicate(self._args)
+        if raw_target._duplicatable:
+            target = raw_target._duplicate(self._args)
 
+            if not self._duplicatable:
+                self.target._unique()
+
+        self.raw_target = raw_target
         self.target = target
 
         renpy.display.render.redraw(self, 0)
@@ -679,6 +715,8 @@ class DynamicImage(renpy.display.core.Displayable):
 
         return True
 
+    _duplicatable = True
+
     def _duplicate(self, args):
 
         if args and args.args:
@@ -691,6 +729,11 @@ class DynamicImage(renpy.display.core.Displayable):
         # same.
         return rv
 
+    def _unique(self):
+        if self.target is not None:
+            self.target._unique()
+            self._duplicatable = False
+
     def _in_current_store(self):
         rv = self._copy()
 
@@ -702,7 +745,7 @@ class DynamicImage(renpy.display.core.Displayable):
 
     def _handles_event(self, event):
         if self.target is None:
-            return False
+            self.find_target()
 
         return self.target._handles_event(event)
 
@@ -790,9 +833,9 @@ class ShownImageInfo(renpy.object.Object):
             self.attributes = { }
             self.shown = set()
 
-            for layer in self.images:
-                for tag in self.images[layer]:
-                    self.attributes[layer, tag] = self.images[layer][tag][1:]
+            for layer in self.images: # type: ignore
+                for tag in self.images[layer]: # type: ignore
+                    self.attributes[layer, tag] = self.images[layer][tag][1:] # type: ignore
                     self.shown.add((layer, tag))
 
     def get_attributes(self, layer, tag, default=()):
@@ -801,8 +844,7 @@ class ShownImageInfo(renpy.object.Object):
         layer.
         """
 
-        if layer is None:
-            layer = renpy.config.tag_layer.get(tag, "master")
+        layer = renpy.exports.default_layer(layer, tag)
 
         return self.attributes.get((layer, tag), default)
 
@@ -815,8 +857,7 @@ class ShownImageInfo(renpy.object.Object):
         tag = name[0]
         rest = name[1:]
 
-        if layer is None:
-            layer = renpy.config.tag_layer.get(tag, "master")
+        layer = renpy.exports.default_layer(layer, tag)
 
         if (layer, tag) not in self.shown:
             return None
@@ -858,7 +899,7 @@ class ShownImageInfo(renpy.object.Object):
         if layer is None:
             layer = 'master'
 
-        for l, t in self.attributes.keys():
+        for l, t in list(self.attributes.keys()):
             if l == layer:
                 del self.attributes[l, t]
 
@@ -876,8 +917,7 @@ class ShownImageInfo(renpy.object.Object):
         tag = name[0]
         rest = name[1:]
 
-        if layer is None:
-            layer = renpy.config.tag_layer.get(tag, "master")
+        layer = renpy.exports.default_layer(layer, tag)
 
         self.attributes[layer, tag] = rest
 
@@ -887,27 +927,29 @@ class ShownImageInfo(renpy.object.Object):
     def predict_hide(self, layer, name):
         tag = name[0]
 
-        if layer is None:
-            layer = renpy.config.tag_layer.get(tag, "master")
+        layer = renpy.exports.default_layer(layer, tag)
 
         if (layer, tag) in self.attributes:
             del self.attributes[layer, tag]
 
         self.shown.discard((layer, tag))
 
-    def apply_attributes(self, layer, tag, name, wanted=[], remove=[]):
+    def apply_attributes(self, layer, tag, name):
         """
         Given a layer, tag, and an image name (with attributes),
         returns the canonical name of an image, if one exists. Raises
-        an exception if it's ambiguious, and returns None if an image
+        an exception if it's ambiguous, and returns None if an image
         with that name couldn't be found.
         """
 
-        if layer is None:
-            layer = renpy.config.tag_layer.get(tag, "master")
+        f = renpy.config.adjust_attributes.get(name[0], None) or renpy.config.adjust_attributes.get(None, None)
+        if f is not None:
+            name = f(name)
+
+        layer = renpy.exports.default_layer(layer, tag)
 
         # If the name matches one that exactly exists, return it.
-        if (name in images) and not (wanted or remove):
+        if name in images:
             ca = getattr(images[name], "_choose_attributes", None)
 
             if ca is None:
@@ -915,21 +957,35 @@ class ShownImageInfo(renpy.object.Object):
 
         nametag = name[0]
 
-        # The set of attributes a matching image may have.
-        optional = set(wanted) | set(self.attributes.get((layer, tag), [ ]))
+        # Find any attributes applied previously.
+        defaults = self.attributes.get((layer, tag), None)
+
+        # If no record, it's the first show, so try to fetch defaults.
+        if defaults is None:
+            f = renpy.config.default_attribute_callbacks.get(name[0], None) \
+                or renpy.config.default_attribute_callbacks.get(None, None)
+            if f is not None:
+                defaults = f(name)
+
+        # The list of attributes a matching image may have.
+        optional = list(defaults) if defaults else [ ]
+
+        # The list of attributes a matching image must have.
+        required = [ ]
 
-        # The set of attributes a matching image must have/not have.
-        # Evaluated in order.
-        required = set()
         for i in name[1:]:
             if i[0] == "-":
-                optional.discard(i[1:])
-                required.discard(i[1:])
-            else:
-                required.add(i)
 
-        for i in remove:
-            optional.discard(i)
+                i = i[1:]
+
+                if i in optional:
+                    optional.remove(i)
+
+                if i in required:
+                    required.remove(i)
+
+            else:
+                required.append(i)
 
         return self.choose_image(nametag, required, optional, name)
 
@@ -943,11 +999,15 @@ class ShownImageInfo(renpy.object.Object):
 
         for attrs, d in image_attributes[tag].items():
 
+            if not all((i in required) or (i in optional) for i in attrs):
+                continue
+
             ca = getattr(d, "_choose_attributes", None)
 
             if ca:
-                ca_required = { i for i in required if i not in attrs }
-                ca_optional = { i for i in optional if i not in attrs }
+                ca_required = [ i for i in required if i not in attrs ]
+                ca_optional = [ i for i in optional if i not in attrs if i not in required ]
+
                 newattrs = ca(tag, ca_required, ca_optional)
 
                 if newattrs is None:
@@ -955,34 +1015,31 @@ class ShownImageInfo(renpy.object.Object):
 
                 attrs = attrs + newattrs
 
-            num_required = 0
-
-            for i in attrs:
-                if i in required:
-                    num_required += 1
-                    continue
-
-                elif i not in optional:
-                    break
-
             else:
 
+                num_required = 0
+
+                for i in attrs:
+                    if i in required:
+                        num_required += 1
+                        continue
+
                 # We don't have any not-found attributes. But we might not
                 # have all of the attributes.
 
                 if num_required != len(required):
                     continue
 
-                len_attrs = len(set(attrs))
+            len_attrs = len(set(attrs))
 
-                if len_attrs < max_len:
-                    continue
+            if len_attrs < max_len:
+                continue
 
-                if len_attrs > max_len:
-                    max_len = len_attrs
-                    matches = [ ]
+            if len_attrs > max_len:
+                max_len = len_attrs
+                matches = [ ]
 
-                matches.append((tag, ) + attrs)
+            matches.append((tag,) + attrs)
 
         if matches is None:
             return None
@@ -996,8 +1053,7 @@ class ShownImageInfo(renpy.object.Object):
             return None
 
 
-renpy.display.core.ImagePredictInfo = ShownImageInfo
-
+renpy.display.core.ImagePredictInfo = ShownImageInfo # type: ignore
 
 # Functions that have moved from this module to other modules,
 # that live here for the purpose of backward-compatibility.
